# type: ignore

from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Set, Tuple, Union

import pytest

from jsondaora import asdataclass, dataclass_asjson, jsondaora
from jsondaora.exceptions import DeserializationError


def tests_should_deserialize_optional_args():
    @dataclass
    class FakeDataclass:
        test: str
        test_default: Optional[int] = None

    dataclass_ = asdataclass(
        {'test': 'test', 'test_default': '1'}, FakeDataclass
    )

    assert dataclass_ == FakeDataclass('test', 1)


def tests_should_deserialize_union_args():
    @dataclass
    class FakeDataclass:
        test_union: Union[int, str]

    dataclass_ = asdataclass({'test_union': b'1'}, FakeDataclass)

    assert dataclass_ == FakeDataclass(1)


def tests_should_deserialize_list_args():
    @dataclass
    class FakeDataclass:
        test_list: List[int]

    dataclass_ = asdataclass({'test_list': [b'1', '2', 3]}, FakeDataclass)

    assert dataclass_ == FakeDataclass([1, 2, 3])


def tests_should_deserialize_list_args_nested():
    @dataclass
    class FakeDataclass:
        test: str

    @dataclass
    class FakeDataclass2:
        fakes: List[FakeDataclass]
        fakeint: int

    fakes_data = [{'test': 'fake11'}, {'test': 'fake12'}, {'test': 'fake13'}]
    dataclass_ = asdataclass(
        {'fakes': fakes_data, 'fakeint': '1'}, FakeDataclass2
    )

    assert dataclass_ == FakeDataclass2(
        [
            FakeDataclass('fake11'),
            FakeDataclass('fake12'),
            FakeDataclass('fake13'),
        ],
        1,
    )


def tests_should_deserialize_tuple_args():
    @dataclass
    class FakeDataclass:
        test_tuple: Tuple[int, ...]

    dataclass_ = asdataclass({'test_tuple': [b'1', '2', 3]}, FakeDataclass)

    assert dataclass_ == FakeDataclass((1, 2, 3))


def tests_should_deserialize_tuple_args_nested():
    @dataclass
    class FakeDataclass:
        test: str

    @dataclass
    class FakeDataclass2:
        fakes: Tuple[FakeDataclass, ...]
        fakeint: int

    fakes_data = [{'test': 'fake11'}, {'test': 'fake12'}, {'test': 'fake13'}]
    dataclass_ = asdataclass(
        {'fakes': fakes_data, 'fakeint': '1'}, FakeDataclass2
    )

    assert dataclass_ == FakeDataclass2(
        (
            FakeDataclass('fake11'),
            FakeDataclass('fake12'),
            FakeDataclass('fake13'),
        ),
        1,
    )


def tests_should_deserialize_tuple_args_limited():
    @dataclass
    class FakeDataclass:
        test_tuple: Tuple[int, str]

    dataclass_ = asdataclass({'test_tuple': [b'1', 2]}, FakeDataclass)

    assert dataclass_ == FakeDataclass((1, '2'))


def tests_should_deserialize_tuple_args_nested_limited():
    @dataclass
    class FakeDataclass:
        test: str

    @dataclass
    class FakeDataclass2:
        fakes: Tuple[FakeDataclass, int]
        fakeint: int

    fakes_data = [{'test': 'fake11'}, '1']
    dataclass_ = asdataclass(
        {'fakes': fakes_data, 'fakeint': '1'}, FakeDataclass2
    )

    assert dataclass_ == FakeDataclass2((FakeDataclass('fake11'), 1), 1,)


def tests_should_raise_error_when_deserializing_invalid_tuple_size():
    @dataclass
    class FakeDataclass:
        test: str

    @dataclass
    class FakeDataclass2:
        fakes: Tuple[FakeDataclass, int]
        fakeint: int

    fakes_data = [{'test': 'fake11'}, '1', None]

    with pytest.raises(DeserializationError) as exc_info:
        asdataclass({'fakes': fakes_data, 'fakeint': '1'}, FakeDataclass2)

    assert exc_info.value.args == (
        f'Invalid type={Tuple[FakeDataclass, int]} for field=fakes',
    )


def tests_should_deserialize_set_args():
    @dataclass
    class FakeDataclass:
        test_set: Set[int]

    dataclass_ = asdataclass({'test_set': [b'1', '2', 3]}, FakeDataclass)

    assert dataclass_ == FakeDataclass({1, 2, 3})


def tests_should_deserialize_set_args_nested():
    @dataclass(unsafe_hash=True)
    class FakeDataclass:
        test: str

    @dataclass
    class FakeDataclass2:
        fakes: Set[FakeDataclass]
        fakeint: int

    fakes_data = [{'test': 'fake11'}, {'test': 'fake12'}, {'test': 'fake13'}]
    dataclass_ = asdataclass(
        {'fakes': fakes_data, 'fakeint': '1'}, FakeDataclass2
    )

    assert dataclass_ == FakeDataclass2(
        {
            FakeDataclass('fake11'),
            FakeDataclass('fake12'),
            FakeDataclass('fake13'),
        },
        1,
    )


def tests_should_deserialize_dict_args():
    @dataclass
    class FakeDataclass:
        test_dict: Dict[int, str]

    dataclass_ = asdataclass(
        {'test_dict': {b'1': b'1', '2': '2', 3: 3}}, FakeDataclass
    )

    assert dataclass_ == FakeDataclass({1: '1', 2: '2', 3: '3'})


def tests_should_deserialize_dict_args_nested():
    @dataclass
    class FakeDataclass:
        test: str

    @dataclass
    class FakeDataclass2:
        fakes: Dict[int, List[FakeDataclass]]
        fakeint: int

    fakes_data = {
        b'1': [{'test': 'fake11'}, {'test': 'fake12'}, {'test': 'fake13'}]
    }
    dataclass_ = asdataclass(
        {'fakes': fakes_data, 'fakeint': '1'}, FakeDataclass2
    )

    assert dataclass_ == FakeDataclass2(
        {
            1: [
                FakeDataclass('fake11'),
                FakeDataclass('fake12'),
                FakeDataclass('fake13'),
            ]
        },
        1,
    )


def tests_should_deserialize_any():
    @dataclass
    class FakeDataclass:
        test_any: Any

    dataclass1 = asdataclass({'test_any': 0.1}, FakeDataclass)
    any_object = object()
    dataclass2 = asdataclass({'test_any': any_object}, FakeDataclass)

    assert dataclass1 == FakeDataclass(0.1)
    assert dataclass2 == FakeDataclass(any_object)


def tests_should_deserialize_any_nested():
    @dataclass
    class FakeDataclass:
        test: str

    @dataclass
    class FakeDataclass2:
        fakes: Set[Any]
        fakeint: int

    any_object = object()
    fakes_data = [any_object, 0.1]
    dataclass_ = asdataclass(
        {'fakes': fakes_data, 'fakeint': '1'}, FakeDataclass2
    )

    assert dataclass_ == FakeDataclass2({any_object, 0.1}, 1,)


def tests_should_deserialize_bytes_to_string():
    @dataclass
    class FakeDataclass:
        test_union: str

    dataclass_ = asdataclass({'test_union': b'test'}, FakeDataclass)

    assert dataclass_ == FakeDataclass('test')


def tests_should_deserialize_nested_jsondict():
    @dataclass
    class FakeDataclass:
        test: str

    @dataclass
    class FakeDataclass2:
        fake: FakeDataclass

    dataclass_ = asdataclass({'fake': {'test': b'test'}}, FakeDataclass2)

    assert dataclass_ == FakeDataclass2(FakeDataclass('test'))


def tests_should_choose_fields_to_deserialize():
    @jsondaora(deserialize_fields=('test2',))
    @dataclass
    class FakeDataclass:
        test: int
        test2: str

    dataclass_ = asdataclass({'test': '1', 'test2': 2}, FakeDataclass)

    assert dataclass_ == FakeDataclass('1', '2')


def tests_should_serialize_all_fields_with_choosen_deserialize_fields():
    @jsondaora(deserialize_fields=('test2',))
    @dataclass
    class FakeDataclass:
        test: int
        test2: str

    dataclass_ = asdataclass({'test': '1', 'test2': 2}, FakeDataclass)

    assert dataclass_asjson(dataclass_) == b'{"test":"1","test2":"2"}'



# type: ignore

import dataclasses
from typing import Any, Dict, List, Optional, Set, Tuple, TypedDict, Union

import pytest

from jsondaora import (
    as_typed_dict,
    as_typed_dict_field,
    jsondaora,
    typed_dict_asjson,
)
from jsondaora.exceptions import DeserializationError


def tests_should_deserialize_optional_args():
    @jsondaora
    class FakeTypedDict(TypedDict):
        test: str
        test_default: Optional[int] = None

    typed_dict = as_typed_dict(
        {'test': 'test', 'test_default': '1'}, FakeTypedDict
    )

    assert typed_dict == {'test': 'test', 'test_default': 1}


def tests_should_deserialize_union_args():
    @jsondaora
    class FakeTypedDict(TypedDict):
        test_union: Union[int, str]

    typed_dict = as_typed_dict({'test_union': b'1'}, FakeTypedDict)

    assert typed_dict == {'test_union': 1}


def tests_should_deserialize_list_args():
    @jsondaora
    class FakeTypedDict(TypedDict):
        test_list: List[int]

    typed_dict = as_typed_dict({'test_list': [b'1', '2', 3]}, FakeTypedDict)

    assert typed_dict == {'test_list': [1, 2, 3]}


def tests_should_deserialize_list_args_nested():
    @jsondaora
    class FakeTypedDict(TypedDict):
        test: str

    @jsondaora
    class FakeTypedDict2(TypedDict):
        fakes: List[FakeTypedDict]
        fakeint: int

    fakes_data = [{'test': 'fake11'}, {'test': 'fake12'}, {'test': 'fake13'}]
    typed_dict = as_typed_dict(
        {'fakes': fakes_data, 'fakeint': '1'}, FakeTypedDict2
    )

    assert typed_dict == {'fakes': fakes_data, 'fakeint': 1}


def tests_should_deserialize_bytes_to_string():
    @jsondaora
    class FakeTypedDict(TypedDict):
        test_union: str

    typed_dict = as_typed_dict({'test_union': b'test'}, FakeTypedDict)

    assert typed_dict == {'test_union': 'test'}


def tests_should_deserialize_nested_jsondict():
    @jsondaora
    class FakeTypedDict(TypedDict):
        test: str

    @jsondaora
    class FakeTypedDict2(TypedDict):
        fake: FakeTypedDict

    typed_dict = as_typed_dict({'fake': {'test': b'test'}}, FakeTypedDict2)

    assert typed_dict == {'fake': {'test': 'test'}}


def tests_should_choose_fields_to_deserialize():
    @jsondaora(deserialize_fields=('test2',))
    class FakeTypedDict(TypedDict):
        test: int
        test2: str

    typed_dict = as_typed_dict({'test': '1', 'test2': 2}, FakeTypedDict)

    assert typed_dict == {'test': '1', 'test2': '2'}


def tests_should_serialize_all_fields_with_choosen_deserialize_fields():
    @jsondaora(deserialize_fields=('test2',))
    class FakeTypedDict(TypedDict):
        test: int
        test2: str

    typed_dict = as_typed_dict({'test': '1', 'test2': 2}, FakeTypedDict)

    assert typed_dict_asjson(typed_dict, FakeTypedDict) in [
        b'{"test":"1","test2":"2"}',
        b'{"test2":"2","test":"1"}',
    ]


def tests_should_set_dataclass_fields_on_typed_dict():
    @jsondaora
    class FakeTypedDict(TypedDict):
        test: int
        test2: str

    fields = dataclasses.fields(FakeTypedDict)

    assert len(fields) == 2
    assert fields[0].type == int
    assert fields[1].type == str


def tests_should_set_dataclass_fields_on_typed_dict_with_inheritance():
    @jsondaora
    class FakeTypedDict2(TypedDict):
        test: float

    @jsondaora
    class FakeTypedDict(FakeTypedDict2):
        test2: int
        test3: str

    @jsondaora
    class Fake:
        fake: FakeTypedDict

    fake = Fake(FakeTypedDict2())

    fake_field = type(fake).__dataclass_fields__[  # type: ignore
        'fake'
    ]
    assert fake_field.type is FakeTypedDict


def tests_should_deserialize_list_args_nested_as_list():
    @jsondaora
    class FakeTypedDict(TypedDict):
        test: str

    @jsondaora
    class FakeTypedDict2(TypedDict):
        fakes: List[FakeTypedDict]
        fakeint: int

    fakes_data = [{'test': 'fake11'}, {'test': 'fake12'}, {'test': 'fake13'}]
    typed_dict_list = as_typed_dict_field(
        [{'fakes': fakes_data, 'fakeint': '1'}], 'list', List[FakeTypedDict2]
    )

    assert typed_dict_list == [{'fakes': fakes_data, 'fakeint': 1}]


def tests_should_deserialize_tuple_args_limited():
    @jsondaora
    class FakeTypedDict(TypedDict):
        test_tuple: Tuple[int, str]

    typed_dict_tuple = as_typed_dict({'test_tuple': [b'1', 2]}, FakeTypedDict)

    assert typed_dict_tuple == {'test_tuple': (1, '2')}


def tests_should_deserialize_tuple_args_nested_limited():
    @jsondaora
    class FakeTypedDict(TypedDict):
        test: str

    @jsondaora
    class FakeTypedDict2(TypedDict):
        fakes: Tuple[FakeTypedDict, int]
        fakeint: int

    fakes_data = [{'test': 'fake11'}, '1']
    typed_dict = as_typed_dict(
        {'fakes': fakes_data, 'fakeint': '1'}, FakeTypedDict2
    )

    assert typed_dict == {
        'fakes': ({'test': 'fake11'}, 1),
        'fakeint': 1,
    }


def tests_should_raise_error_when_deserializing_invalid_tuple_size():
    @jsondaora
    class FakeTypedDict(TypedDict):
        test: str

    @jsondaora
    class FakeTypedDict2(TypedDict):
        fakes: Tuple[FakeTypedDict, int]
        fakeint: int

    fakes_data = [{'test': 'fake11'}, '1', None]

    with pytest.raises(DeserializationError) as exc_info:
        as_typed_dict({'fakes': fakes_data, 'fakeint': '1'}, FakeTypedDict2)

    assert exc_info.value.args == (
        f'Invalid type={Tuple[FakeTypedDict, int]} for field=fakes',
    )


def tests_should_deserialize_set_args():
    @jsondaora
    class FakeTypedDict(TypedDict):
        test_set: Set[int]

    typed_dict = as_typed_dict({'test_set': [b'1', '2', 3]}, FakeTypedDict)

    assert typed_dict == {'test_set': {1, 2, 3}}


def tests_should_deserialize_set_args_nested():
    @jsondaora
    class FakeTypedDict2(TypedDict):
        fakes: Set[Tuple[int, str, float]]
        fakeint: int

    fakes_data = [[1, 2, 3]]
    typed_dict = as_typed_dict(
        {'fakes': fakes_data, 'fakeint': '1'}, FakeTypedDict2
    )

    assert typed_dict == {'fakes': {(1, '2', 3.0)}, 'fakeint': 1}


def tests_should_deserialize_dict_args():
    @jsondaora
    class FakeTypedDict(TypedDict):
        test_dict: Dict[int, str]

    typed_dict = as_typed_dict(
        {'test_dict': {b'1': b'1', '2': '2', 3: 3}}, FakeTypedDict
    )

    assert typed_dict == {'test_dict': {1: '1', 2: '2', 3: '3'}}


def tests_should_deserialize_dict_args_nested():
    @jsondaora
    class FakeTypedDict(TypedDict):
        test: str

    @jsondaora
    class FakeTypedDict2(TypedDict):
        fakes: Dict[int, List[FakeTypedDict]]
        fakeint: int

    fakes_data = {
        b'1': [{'test': 'fake11'}, {'test': 'fake12'}, {'test': 'fake13'}]
    }
    typed_dict = as_typed_dict(
        {'fakes': fakes_data, 'fakeint': '1'}, FakeTypedDict2
    )

    assert typed_dict == FakeTypedDict2(
        {
            'fakes': {
                1: [{'test': 'fake11'}, {'test': 'fake12'}, {'test': 'fake13'}]
            },
            'fakeint': 1,
        },
    )


def tests_should_deserialize_any():
    @jsondaora
    class FakeTypedDict(TypedDict):
        test_any: Any

    typed_dict1 = as_typed_dict({'test_any': 0.1}, FakeTypedDict)
    any_object = object()
    typed_dict2 = as_typed_dict({'test_any': any_object}, FakeTypedDict)

    assert typed_dict1 == {'test_any': 0.1}
    assert typed_dict2 == {'test_any': any_object}


def tests_should_deserialize_any_nested():
    @jsondaora
    class FakeTypedDict(TypedDict):
        test: str

    @jsondaora
    class FakeTypedDict2(TypedDict):
        fakes: Set[Any]
        fakeint: int

    any_object = object()
    fakes_data = [any_object, 0.1]
    typed_dict = as_typed_dict(
        {'fakes': fakes_data, 'fakeint': '1'}, FakeTypedDict2
    )

    assert typed_dict == {
        'fakes': {any_object, 0.1},
        'fakeint': 1,
    }



import dataclasses
from enum import Enum
from typing import Any, Dict, Set, Type

from .deserializers import deserialize_jsondict_fields
from .exceptions import DeserializationError
from .fields import SerializeFields


def asdataclass(
    jsondict: Dict[str, Any], cls: Type[Any], skip_fields: Set[str] = set()
) -> Any:
    kwargs = deserialize_jsondict_fields(jsondict, cls, skip_fields)

    try:
        return cls(**kwargs)
    except TypeError as error:
        raise DeserializationError(
            cls.__name__, jsondict, error, cls
        ) from error


def asdict(instance: Any) -> Dict[str, Any]:
    fields = SerializeFields.get_fields(type(instance))

    if not fields:
        fields = dataclasses.fields(instance)

    return {
        f.name: (
            asdict(value)
            if dataclasses.is_dataclass(
                value := getattr(instance, f.name, f.default)  # noqa
            )
            else (
                (
                    asdict(enum_value)
                    if dataclasses.is_dataclass(
                        enum_value := value.value  # noqa
                    )
                    else enum_value
                )
                if isinstance(value, Enum)
                else value
            )
        )
        for f in fields
    }
